"""
Utilities for creating QC JSON files (https://molssi-qc-schema.readthedocs.io/en/latest/index.html)
"""
from dataclasses import dataclass, field
from typing import Literal, Optional

import dataclasses_json

QCMethods = Literal["RHF", "RIMP2"]

QCBasisSet = Literal[
    "STO-2G",
    "STO-3G",
    "STO-4G",
    "STO-5G",
    "STO-6G",
    "3-21G",
    "4-31G",
    "5-21G",
    "6-21G",
    "6-31G",
    "6-31G*",
    "6-31G**",
    "6-31+G",
    "6-31+G*",
    "6-31+G**",
    "6-31++G",
    "6-31++G*",
    "6-31++G**",
    "6-31G(2df,p)",
    "6-31G(3df,3pd)",
    "6-31G(d,p)",
    "6-311G",
    "6-311G*",
    "6-311G**",
    "6-311+G",
    "6-311+G*",
    "6-311+G**",
    "6-311++G",
    "6-311++G*",
    "6-311++G**",
    "6-311G(d,p)",
    "6-311G(2df,2pd)",
    "6-311+G(2d,p)",
    "6-311++G(2d,2p)",
    "6-311++G(3df,3pd)",
    "PCSeg-0",
    "PCSeg-1",
    "PCSeg-2",
    "cc-pVDZ-RIFIT",
    "aug-cc-pVDZ-RIFIT",
    "cc-pVTZ-RIFIT",
    "aug-cc-pVTZ-RIFIT",
    "cc-pVDZ",
    "aug-cc-pVDZ",
    "cc-pVTZ",
    "aug-cc-pVTZ",
]


class DataClassJsonMixin(dataclasses_json.DataClassJsonMixin):
    """Override dataclass mixin so that we don't have `"property": null,`s in our output"""

    dataclass_json_config = dataclasses_json.config(  # type: ignore
        undefined=dataclasses_json.Undefined.EXCLUDE,
        exclude=lambda f: f is None,  # type: ignore
    )["dataclasses_json"]


@dataclass
class QCJSONFragments(DataClassJsonMixin):
    """ "fragments" in qc json"""

    fragid: list[int] = field(default_factory=list)
    nfrag: int = 0
    broken_bonds: list[int] = field(default_factory=list)
    fragment_charges: list[int] = field(default_factory=list)
    frag_size: list[int] = field(default_factory=list)


@dataclass
class QCMol(DataClassJsonMixin):
    """ "molecule" in qc json"""

    geometry: list[float] = field(default_factory=list)
    symbols: list[str] = field(default_factory=list)
    fragments: QCJSONFragments = QCJSONFragments()


@dataclass
class QCModel(DataClassJsonMixin):
    """ "model" in qc json"""

    method: QCMethods = "RHF"
    fragmentation: bool = False
    basis: QCBasisSet = "6-31G*"
    aux_basis: str = "cc-pVDZ-RIFIT"


@dataclass
class QCSCF(DataClassJsonMixin):
    """ "keywords.scf" in qc json"""

    niter: int = 40
    ndiis: int = 12
    scf_conv: float = 1e-6
    dynamic_screening_threshold_exp: int = 10
    debug: bool | None = None
    convergence_metric: str = "diis"


@dataclass
class QCFrag(DataClassJsonMixin):
    """ "keywords.frag" in qc json"""

    reference_fragment: Optional[int] = None
    method: str = "MBE"
    fragmentation_level: int = 2
    ngpus_per_node: int = 4
    subset_of_fragments_to_calculate: Optional[list[int]] = None
    monomer_cutoff: int = 40
    monomer_mp2_cutoff: int = 40
    dimer_cutoff: int = 40
    dimer_mp2_cutoff: int = 40
    trimer_mp2_cutoff: int = 20
    lattice_energy_calc: bool = False
    trimer_cutoff: int = 30


@dataclass
class QCKeywords(DataClassJsonMixin):
    """ "keywords" in qc json"""

    scf: QCSCF = QCSCF()
    frag: QCFrag = QCFrag()


@dataclass
class QDXQCMol(DataClassJsonMixin):
    """ "molecule" in qc json"""

    geometry: list[float] = field(default_factory=list)
    symbols: list[str] = field(default_factory=list)
    fragments: Optional[list[list[int]]] = field(default_factory=list)
    connectivity: Optional[list[tuple[int, int, int]]] = field(default_factory=list)
    fragment_charges: Optional[list[int]] = field(default_factory=list)


@dataclass
class QDXQCInput(DataClassJsonMixin):
    """Final qc json input file"""

    topology: QDXQCMol = QDXQCMol()
    model: QCModel = QCModel()
    keywords: QCKeywords = QCKeywords()
    driver: str = "energy"


@dataclass
class QDXEnergy(DataClassJsonMixin):
    energy_type: Literal["lattice", "SPE"] | None  # Whether the calculation is lattice against a ref. monomer
    hf_total: float  # Final single-point energy (HF)
    mp_ss_total: float  # Final single-point energy correction (MP2 same-spin)
    mp_os_total: float  # Final single-point energy correction (MP2 opposite-spin)

    hf: list[float]  # HF energy per fragment group
    mp_ss: list[float]  # MP2 same-spin corrections per fragment group
    mp_os: list[float]  # MP2 opposite-spin corrections per fragment group
    n_mers: list[list[int]]  # Fragment groups
    n_mer_distances: list[float] | None  # Distances between fragments



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tengu']

package_data = \
{'': ['*']}

install_requires = \
['dataclasses-json>=0.5.7,<0.6.0',
 'datargs>=0.11.0,<0.12.0',
 'gql[requests]>=3.4.0,<4.0.0',
 'rdkit-pypi>=2022.9.5,<2023.0.0']

setup_kwargs = {
    'name': 'tengu-py',
    'version': '0.4.0',
    'description': 'Python SDK for interacting with the QDX Tengu API',
    'long_description': '# Tengu-py: Python SDK for the QDX Tengu API\n\nThis package exposes a simple provider and CLI for the different tools exposed by the QDX Tengu GraphQL API.\n\n## Usage\n\n### As a library\n\n``` python\nimport json\nfrom pathlib import Path\n\nimport tengu\n\nTOKEN = "your qdx access token"\n\n# get our client to talk with the API\nclient = tengu.Provider(access_token=TOKEN)\n\n# get modules that can be run\nclient.modules()\n\n## running convert\n\n# path to protein pdb with correct charges and protonation\nprotein_pdb = Path("./examples/4w9f_prepared_protein.pdb")\n\n# get base64 encoded data\n\nfile_arg = provider.upload_arg(protein_pdb)\n\nres = client.run("github:talo/tengu-prelude/f8e2e55d9bd428aa7f2bbe3f87c24775fa592b10#convert", [ \n{ "value": "PDB" }, file_arg\n])\n\n// res contains "id" - the instance id; and "outs" - the ids of the return values \n\n// we can pass arguments by "id" reference or by value literal\n\nclient.run("github:talo/tengu-prelude/f8e2e55d9bd428aa7f2bbe3f87c24775fa592b10#pick_conformer", [ \n{ "id": res["outs"][0]["id"] }, { "value": 1 }\n])\n\nclient.poll_module_instance(id) \n// status, progress, logs, outs - out values will be null until module_instance is done\n```\n\n## Sample QP Run\n\n``` python\nfrag_keywords = {\n    "dimer_cutoff": 25,\n    "dimer_mp2_cutoff": 25,\n    "fragmentation_level": 2,\n    "method": "MBE",\n    "monomer_cutoff": 30,\n    "monomer_mp2_cutoff": 30,\n    "ngpus_per_node": 4,\n    "reference_fragment": 293,\n    "trimer_cutoff": 10,\n    "trimer_mp2_cutoff": 10,\n    "lattice_energy_calc": True,\n}\n\nscf_keywords = {\n    "convergence_metric": "diis",\n    "dynamic_screening_threshold_exp": 10,\n    "ndiis": 8,\n    "niter": 40,\n    "scf_conv": 0.000001,\n}\n\ndefault_model = {"method": "RIMP2", "basis": "cc-pVDZ", "aux_basis": "cc-pVDZ-RIFIT", "frag_enabled": True}\n\nqp_instances = client.qp_run(\n    "github:talo/tengu-prelude/0986e4b23780d5e976e7938dc02a949185090fa1#qp_gen_inputs",\n    "github:talo/tengu-prelude/0986e4b23780d5e976e7938dc02a949185090fa1#hermes_energy",\n    "github:talo/tengu-prelude/0986e4b23780d5e976e7938dc02a949185090fa1#qp_collate",\n    provider.upload_arg(Path("some.pdb")),\n    provider.upload_arg(Path("some.gro")),\n    provider.upload_arg(Path("some.sdf")),\n    Arg(None, "sdf"),\n    Arg(None, "MOL"),\n    Arg(\n        None,\n        default_model,\n    ),\n    Arg(None, {"frag": frag_keywords, "scf": scf_keywords}),\n    Arg(\n        None,\n        [\n            ("GLY", 993),\n            ("ASP", 994),\n            ("VAL", 863),\n            ("LYS", 882),\n            ("TYR", 931),\n            ("GLY", 935),\n            ("VAL", 911),\n            ("GLU", 930),\n            ("ALA", 880),\n            ("LEU", 983),\n            ("PRO", 933),\n            ("LEU", 855),\n            ("MET", 929),\n            ("SER", 936),\n            ("LEU", 932),\n        ],\n    ),\n    "GADI",\n    {"walltime": 420},\n    autopoll = (10, 100) # optionally configure polling to wait on the final instance, \n                         # and clean up if any of the prior instances fails\n)\n\n# if you set autpoll, you will get the results of the qp_collate instance,\n# otherwise you will get an array with all the spawned instances, and have to poll manually\nclient.poll_module_instance(qp_collate_instance[2]["id"]) \n```\n',
    'author': 'Ryan Swart',
    'author_email': 'ryan@talosystems.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python3

import json
import sys

import numpy as np
from rdkit import Chem
from rdkit.Chem import AllChem, Descriptors

from .data import QDXQCMol

DEBUG = False


def standard_bond_length(sym):
    # https://www.ibchem.com/IB16/Section04-energetics/data/Hbde_data.htm
    bonds = {
        "H": 0.74,
        "C": 1.09,
        "N": 1.01,
        "O": 0.96,
        "F": 0.92,
        "P": 1.44,
        "S": 1.34,
        "CL": 1.27,
        "SE": 1.46,
    }
    return bonds.get(sym.upper(), 1.0)


def scale_bond(A, B, d):
    R = np.subtract(B, A)
    r = np.linalg.norm(R)
    return np.add(A, R * d / r)


def nuclear_charge(frag):
    Ztot = 0
    for atom in frag.GetAtoms():
        Ztot += atom.GetAtomicNum()
    return Ztot


def dump_to_file(frags, fmt="mol"):
    for i in range(len(frags)):
        frag = frags[i]
        fname = "frag{}.{}".format(i, fmt)
        with open(fname, "w") as file:
            if fmt == "mol":
                file.write(Chem.MolToMolBlock(frag))
            else:
                file.write(Chem.MolToXYZBlock(frag))


def extract_fragments(data: QDXQCMol):
    """
    Takes in JSON formatted data and extracts the fragments into Chem.RWMol objects.
    These fragments have the capping hydrogen atoms added with standardized bonds.
    """
    symbols = data.symbols
    fragments = data.fragments or []
    geometry = data.geometry
    connectivity = data.connectivity or []

    coords = np.array(geometry)
    coords = coords.reshape(len(symbols), 3)

    frags = []
    atom2frag = {}
    atom2atom = {}

    # Create XYZ block to parse as a RWMol
    for ifrag in range(len(fragments)):
        xyz = "{}\n\n".format(len(fragments[ifrag]))
        iatom = 0

        for atom in fragments[ifrag]:
            line = "{}   {}   {}   {}\n".format(symbols[atom], *coords[atom])
            xyz += line
            atom2frag[atom] = ifrag
            atom2atom[atom] = iatom
            iatom += 1

        mol = Chem.RWMol(Chem.MolFromXYZBlock(xyz))
        frags.append(mol)

    # Assign bond to their respective fragments, adding caps for inter-fragment bonds
    for bond in connectivity:
        atom1, atom2, order = bond
        f1 = atom2frag[atom1]
        f2 = atom2frag[atom2]
        a1 = atom2atom[atom1]
        a2 = atom2atom[atom2]

        if f1 == f2:
            # intra-fragment bond
            if order == 1:
                frags[f1].AddBond(a1, a2, Chem.BondType.SINGLE)
            elif order == 2:
                frags[f1].AddBond(a1, a2, Chem.BondType.DOUBLE)
            elif order == 3:
                frags[f1].AddBond(a1, a2, Chem.BondType.TRIPLE)
            else:
                sys.stderr.write("WARNING: unknown bond order encountered: {}".format(order))

        else:
            # inter-fragment bond => add H caps
            conf = frags[f1].GetConformer(-1)
            idx = frags[f1].AddAtom(Chem.Atom(1))
            blen = standard_bond_length(symbols[atom1])
            pos = scale_bond(coords[atom1], coords[atom2], blen)
            conf.SetAtomPosition(idx, pos)

            conf = frags[f2].GetConformer(-1)
            idx = frags[f2].AddAtom(Chem.Atom(1))
            blen = standard_bond_length(symbols[atom2])
            pos = scale_bond(coords[atom2], coords[atom1], blen)
            conf.SetAtomPosition(idx, pos)

    # if DEBUG: dump_to_file(frags,"xyz")

    return frags


def calc_formal_charge(frag):
    """
    Takes in a RWMol object and looks for charged functional groups, adding
    formal charges as required.
    """
    frag.UpdatePropertyCache(strict=False)

    # N+ on LYS, ARG and HIS
    for indices in frag.GetSubstructMatches(Chem.MolFromSmarts("N")):
        atom = frag.GetAtomWithIdx(indices[0])
        if atom.GetExplicitValence() == 4 and atom.GetFormalCharge() == 0:
            atom.SetFormalCharge(1)
            if DEBUG:
                print(
                    "N+   {}:  {}   {}   {}".format(
                        indices[0], atom.GetAtomicNum(), atom.GetExplicitValence(), atom.GetFormalCharge()
                    )
                )

    # S- on CYS
    for indices in frag.GetSubstructMatches(Chem.MolFromSmarts("S")):
        atom = frag.GetAtomWithIdx(indices[0])
        if atom.GetExplicitValence() == 1 and atom.GetFormalCharge() == 0:
            atom.SetFormalCharge(-1)
            if DEBUG:
                print(
                    "S-   {}:  {}   {}   {}".format(
                        indices[0], atom.GetAtomicNum(), atom.GetExplicitValence(), atom.GetFormalCharge()
                    )
                )

    # COO- on ASP and GLU
    for indices in frag.GetSubstructMatches(Chem.MolFromSmarts("C(=O)[O]")):
        atom = frag.GetAtomWithIdx(indices[2])
        if atom.GetExplicitValence() == 1 and atom.GetFormalCharge() == 0:
            atom.SetFormalCharge(-1)
            if DEBUG:
                print(
                    "COO- {}:  {}   {}   {}".format(
                        indices[2], atom.GetAtomicNum(), atom.GetExplicitValence(), atom.GetFormalCharge()
                    )
                )

    # Protonated imidazole ring on HIS
    for indices in frag.GetSubstructMatches(Chem.MolFromSmiles("C1=CNCN1")):
        c = frag.GetAtomWithIdx(indices[3])
        if c.GetExplicitValence() == 4:
            continue
        n1 = frag.GetAtomWithIdx(indices[2])
        n2 = frag.GetAtomWithIdx(indices[4])

        if n1.GetExplicitValence() == 3 and n2.GetExplicitValence() == 3:
            if n1.GetFormalCharge() == 0 and n2.GetFormalCharge() == 0:
                n1.SetFormalCharge(1)
            if DEBUG:
                print(
                    "HIS+ {}:  {}   {}   {}".format(
                        indices[2], n1.GetAtomicNum(), n1.GetExplicitValence(), n1.GetFormalCharge()
                    )
                )

    return Chem.rdmolops.GetFormalCharge(frag)


def run(data: QDXQCMol):
    frags = extract_fragments(data)
    # if DEBUG: dump_to_file(frags)

    charges = []

    radical = 0
    ifrag = 0
    for f in frags:
        q = calc_formal_charge(f)
        n = nuclear_charge(f)
        if (n - q) % 2 != 0:
            radical += 1

        if DEBUG:
            print(Chem.MolToSmiles(f))
            print("Formal charge frag {}: {}".format(ifrag, q))
            print("Total nuclear charge:  {}".format(n))
            print("Number of electrons:   {}".format(n - q))
            if (n - q) % 2 != 0:
                print("*** Non singlet system ***")
            print("")

        ifrag += 1
        charges.append(q)

    if radical > 0:
        sys.stderr.write("WARNING: {} radical systems found\n".format(radical))

    if not DEBUG:
        data.fragment_charges = charges
        return data


def main():
    if sys.argv[1] == "-d":
        global DEBUG
        DEBUG = True
        fname = sys.argv[2]
    else:
        fname = sys.argv[1]

    with open(fname, "r") as f:
        data = json.load(f)
        result = run(data)
        print(json.dumps(result, indent=4))


if __name__ == "__main__":
    main()
